@@ -15,11 +15,32 @@ from __future__ import absolute_import, division, unicode_literals
 
 from weakref import ref
 
-from mo_future import allocate_lock as _allocate_lock, text
-from mo_logs import Log
 from mo_dots import is_null
+from mo_future import allocate_lock as _allocate_lock
+from mo_logs import Log, Except
+from mo_logs.exceptions import get_stacktrace
 
 DEBUG = False
+TRACE_THEN = False  # GRAB STACK TRACE OF then() CALL FOR BLAME
+
+
+def standard_warning(cause):
+    Log.warning(
+        "Trigger on Signal.go() failed, and no error function provided!",
+        cause=cause,
+        stack_depth=1,
+    )
+
+
+def debug_warning(stacktrace):
+    def warning(cause):
+        Log.warning(
+            "Trigger on Signal.go() failed, and no error function provided!",
+            cause=[cause, Except(template="attached at", trace=stacktrace)],
+            stack_depth=1,
+        )
+
+    return warning
 
 
 class Signal(object):
@@ -41,9 +62,6 @@ class Signal(object):
         self.job_queue = None
         self.waiting_threads = None
 
-    def __str__(self):
-        return str(self._go)
-
     def __bool__(self):
         return self._go
 
@@ -79,11 +97,11 @@ class Signal(object):
         DEBUG and self._name and Log.note("GO! {{name|quote}}", name=self.name)
 
         if self._go:
-            return
+            return self
 
         with self.lock:
             if self._go:
-                return
+                return self
             self._go = True
 
         DEBUG and self._name and Log.note("internal GO! {{name|quote}}", name=self.name)
@@ -98,47 +116,53 @@ class Signal(object):
                 t.release()
 
         if jobs:
-            for j in jobs:
+            for j, e in jobs:
                 try:
                     j()
                 except Exception as cause:
-                    Log.warning("Trigger on Signal.go() failed!", cause=cause)
+                    e(cause)
+        return self
 
-    def then(self, target):
+    def then(self, target, error=standard_warning):
         """
         RUN target WHEN SIGNALED
         """
-        if not target:
-            Log.error("expecting target")
+        if DEBUG:
+            if not target:
+                Log.error("expecting target")
+            if isinstance(target, Signal):
+                Log.error("expecting a function, not a signal")
 
         with self.lock:
             if not self._go:
-                DEBUG and self._name and Log.note(
-                    "Adding target to signal {{name|quote}}", name=self.name
-                )
+                if TRACE_THEN:
+                    error = debug_warning(get_stacktrace(1))
 
                 if not self.job_queue:
-                    self.job_queue = [target]
+                    self.job_queue = [(target, error)]
                 else:
-                    self.job_queue.append(target)
-                return
+                    self.job_queue.append((target, error))
+                return self
 
-        DEBUG and Log.note(
-            "Signal {{name|quote}} already triggered, running job immediately",
-            name=self.name,
-        )
-        target()
+        try:
+            target()
+        except Exception as cause:
+            error(cause)
+        return self
 
-    def remove_go(self, target):
+    def remove_then(self, target):
         """
         FOR SAVING MEMORY
         """
+        if self._go:
+            return
+
         with self.lock:
-            if not self._go:
-                try:
-                    self.job_queue.remove(target)
-                except ValueError:
-                    pass
+            if not self._go and self.job_queue:
+                for i, (j, e) in enumerate(self.job_queue):
+                    if j == target:
+                        del self.job_queue[i]
+                        break
 
     @property
     def name(self):
@@ -148,10 +172,10 @@ class Signal(object):
             return self._name
 
     def __str__(self):
-        return self.name.decode(text)
+        return f"{self._go} ({self.name})"
 
     def __repr__(self):
-        return text(repr(self._go))
+        return repr(self._go)
 
     def __or__(self, other):
         if is_null(other):
@@ -161,9 +185,7 @@ class Signal(object):
         if self or other:
             return DONE
 
-        output = Signal(self.name + " | " + other.name)
-        OrSignal(output, (self, other))
-        return output
+        return or_signal(self, other)
 
     def __ror__(self, other):
         return self.__or__(other)
@@ -175,9 +197,9 @@ class Signal(object):
             Log.error("Expecting OR with other signal")
 
         if DEBUG and self._name:
-            output = Signal(self.name + " and " + other.name)
+            output = Signal(self.name + " & " + other.name)
         else:
-            output = Signal(self.name + " and " + other.name)
+            output = Signal(self.name + " & " + other.name)
 
         gen = AndSignals(output, 2)
         self.then(gen.done)
@@ -209,6 +231,12 @@ class AndSignals(object):
             self.signal.go()
 
 
+def or_signal(*dependencies):
+    output = Signal(" | ".join(d.name for d in dependencies))
+    OrSignal(output, dependencies)
+    return output
+
+
 class OrSignal(object):
     """
     A SELF-REFERENTIAL CLUSTER OF SIGNALING METHODS TO IMPLEMENT __or__()
@@ -224,16 +252,21 @@ class OrSignal(object):
             d.then(self)
         signal.then(self.cleanup)
 
-    def cleanup(self, r=None):
-        for d in self.dependencies:
-            d.remove_go(self)
-        self.dependencies = []
+    def cleanup(self, _=None):
+        self.dependencies, dependencies = [], self.dependencies
+        for d in dependencies:
+            d.remove_then(self)
 
-    def __call__(self, *args, **kwargs):
-        s = self.signal()
-        if s is not None:
-            s.go()
+    def __call__(self):
+        signal = self.signal()
+        if signal is not None:
+            signal.go()
+
+    def __hash__(self):
+        return id(self)
+
+    def __eq__(self, other):
+        return id(self) == id(other)
 
 
-DONE = Signal()
-DONE.go()
+DONE = Signal().go()
